unit uMain;   

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, IdBaseComponent, IdComponent, IdTCPServer, StdCtrls, ImgList,
  ExtCtrls, Buttons, Spin;
           
const
  sApplicationIniSection='Preferences';

type
  TfmMain = class(TForm)
    Server: TIdTCPServer;
    laPort: TLabel;
    sedPort: TSpinEdit;
    sbtStartStop: TSpeedButton;
    sbtClear: TSpeedButton;
    sbtSaveAs: TSpeedButton;
    laDeveloper: TLabel;
    laName: TLabel;
    laWebSiteAddress: TLabel;
    laMadeInRussia: TLabel;
    imgFlag: TImage;
    ilButtons: TImageList;
    memLog: TMemo;
    laLog: TLabel;
    sdMain: TSaveDialog;
    btUserSettings: TButton;
    Timer: TTimer;
    procedure laWebSiteAddressMouseEnter(Sender: TObject);
    procedure laWebSiteAddressMouseLeave(Sender: TObject);
    procedure laWebSiteAddressClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure sbtClearClick(Sender: TObject);
    procedure sbtSaveAsClick(Sender: TObject);
    procedure btUserSettingsClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ServerConnect(AThread: TIdPeerThread);
    procedure sbtStartStopClick(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure ServerDisconnect(AThread: TIdPeerThread);
    procedure ServeridcUSERCommand(ASender: TIdCommand);
    procedure ServeridcPASSCommand(ASender: TIdCommand);
    procedure ServeridcCWDCommand(ASender: TIdCommand);
    procedure ServerNoCommandHandler(ASender: TIdTCPServer;
      const AData: String; AThread: TIdPeerThread);
    procedure ServeridcPWDCommand(ASender: TIdCommand);
    procedure ServeridcPASVCommand(ASender: TIdCommand);
    procedure ServeridcLISTCommand(ASender: TIdCommand);
    procedure ServeridcSYSTCommand(ASender: TIdCommand);
    procedure ServeridcTYPECommand(ASender: TIdCommand);
    procedure ServeridcRETRCommand(ASender: TIdCommand);
    procedure ServeridcSTORCommand(ASender: TIdCommand);
    procedure ServeridcQUITCommand(ASender: TIdCommand);
  private
    procedure LoadfmMainSettings;
    procedure SavefmMainSettings;
    procedure LoadUserSettings;
    procedure SaveUserSettings;
    procedure EnableControls(bEnable: boolean);
  public
  end;

var
  fmMain: TfmMain;

  sExeDirPath, sIniFilePath: String;

  bAllowAnonymousLogin: boolean;
  sAnonymousDir: String;
  bCheckPassForValidMail: boolean;

  slUsers, slLog: TStringList;

  LogCS: TRTLCriticalSection;

implementation

uses
  ShellAPI, IniFiles, IdSimpleServer, IdIOHandlerSocket, MD5, uUserSettings;
           
{$R *.dfm}
{$R Manifest.res}

type
  TUserType = (utNone, utAnonymous, utRegisteredUser);
  TTransferType = (ttASCII, ttBinary);

threadvar
  UserName: String;
  UserType: TUserType;
  Authenticated: boolean;
  Password: String;
  CurrentDir: String;
  PASV: boolean;
  TransferType: TTransferType;
  DataChannel: TIdSimpleServer;

procedure UnsetROAttrib(const sFileName: String);
var
  dwAttribs: DWORD;
  lpFileName: PChar;
begin
  if FileExists(sFileName) then
    begin
      lpFileName:=PChar(sFileName);
      dwAttribs:=GetFileAttributes(lpFileName);
      if (dwAttribs and FILE_ATTRIBUTE_READONLY) <> 0 then
        SetFileAttributes(lpFileName, dwAttribs xor FILE_ATTRIBUTE_READONLY);
    end;
end;

function GetExeDir: String;
var
  ts, ts2: String;
begin
  GetDir(0, ts);
  ts:=ExcludeTrailingPathDelimiter(ts);
  ts2:=ExcludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
  if UpperCase(ts)=UpperCase(ts2) then
    Result:=ts
  else
    Result:=ts2;
end;

procedure TfmMain.laWebSiteAddressMouseEnter(Sender: TObject);
begin
  (Sender as TLabel).Font.Color:=clRed;
end;

procedure TfmMain.laWebSiteAddressMouseLeave(Sender: TObject);
begin
  (Sender as TLabel).Font.Color:=clBlue;
end;

procedure TfmMain.laWebSiteAddressClick(Sender: TObject);
begin
  ShellExecute(Application.Handle, PChar('open'), PChar('http://www.studforum.ru'), nil, nil, SW_NORMAL);
end;

procedure TfmMain.FormCreate(Sender: TObject);
begin
  slUsers:=TStringList.Create;
  slLog:=TStringList.Create;

  sExeDirPath:=GetExeDir;
  sIniFilePath:=ChangeFileExt(Application.ExeName, '.ini');
  LoadfmMainSettings;
  LoadUserSettings;
  ilButtons.GetBitmap(0, sbtStartStop.Glyph);

  InitializeCriticalSection(LogCS);
end;

procedure TfmMain.sbtClearClick(Sender: TObject);
begin
  memLog.Clear;
  sdMain.FileName:='';
end;

procedure TfmMain.sbtSaveAsClick(Sender: TObject);
begin
  if sdMain.Execute then
    try
      memLog.Lines.SaveToFile(sdMain.FileName);
    except
      // error message here
    end;
end;

procedure TfmMain.btUserSettingsClick(Sender: TObject);
var
  i, Count: Integer;
begin
  fmUserSettings:=TfmUserSettings.Create(Self);
  try
    fmUserSettings.cbAllowAnonymousLogin.Checked:=bAllowAnonymousLogin;
    fmUserSettings.edAnonymousDir.Text:=sAnonymousDir;
    fmUserSettings.cbCheckPassForValidMail.Checked:=bCheckPassForValidMail;
    Count:=slUsers.Count div 3;
    for i:=0 to Count-1 do
      with fmUserSettings.lvUsers.Items.Add do
        begin
          Caption:=slUsers[i*3];
          SubItems.Add(slUsers[i*3+1]);
          SubItems.Add(slUsers[i*3+2]);
        end;
    fmUserSettings.ShowModal;
    if fmUserSettings.ModalResult=mrOk then
      SaveUserSettings;
  finally
    fmUserSettings.Free;
  end;
end;

procedure TfmMain.LoadfmMainSettings;
var
  ifMain: TIniFile;
  WindowPlacement: TWindowPlacement;
  DeltaX, DeltaY: Integer;
begin
  if FileExists(sIniFilePath) then
    begin
      ifMain:=TIniFile.Create(sIniFilePath);
      ifMain.UpdateFile;

      WindowPlacement.length:=SizeOf(WindowPlacement);
      GetWindowPlacement(Handle, @WindowPlacement);
      WindowPlacement.showCmd:=SW_HIDE; {the form isn't visible yet}

      DeltaX:=ifMain.ReadInteger(sApplicationIniSection, 'Left', (Screen.Width-(WindowPlacement.rcNormalPosition.Right-WindowPlacement.rcNormalPosition.Left)) div 2)-WindowPlacement.rcNormalPosition.Left;
      DeltaY:=ifMain.ReadInteger(sApplicationIniSection, 'Top', (Screen.Height-(WindowPlacement.rcNormalPosition.Bottom-WindowPlacement.rcNormalPosition.Top)) div 2)-WindowPlacement.rcNormalPosition.Top;

      Inc(WindowPlacement.rcNormalPosition.Left, DeltaX);
      Inc(WindowPlacement.rcNormalPosition.Right, DeltaX);
      Inc(WindowPlacement.rcNormalPosition.Top, DeltaY);
      Inc(WindowPlacement.rcNormalPosition.Bottom, DeltaY);

      SetWindowPlacement(Handle, @WindowPlacement);

      sedPort.Value:=ifMain.ReadInteger(sApplicationIniSection, 'Port', 8021);

      ifMain.Free;
    end
  else
    begin
      fmMain.Left:=(Screen.Width - fmMain.Width) div 2;
      fmMain.Top:=(Screen.Height - fmMain.Height) div 2;
    end;
end;

procedure TfmMain.SavefmMainSettings;
var
  ifMain: TIniFile;
  WindowPlacement: TWindowPlacement;
begin
  try
    ifMain:=TIniFile.Create(sIniFilePath);

    WindowPlacement.length := SizeOf(WindowPlacement);
    GetWindowPlacement(Handle, @WindowPlacement);

    ifMain.WriteInteger(sApplicationIniSection, 'Left', WindowPlacement.rcNormalPosition.Left);
    ifMain.WriteInteger(sApplicationIniSection, 'Top', WindowPlacement.rcNormalPosition.Top);

    ifMain.WriteInteger(sApplicationIniSection, 'Port', sedPort.Value);

    ifMain.Free;
  except
  end;
end;

procedure TfmMain.FormDestroy(Sender: TObject);
begin
  if sbtStartStop.Tag = 1 then
    sbtStartStop.Click;
           
  DeleteCriticalSection(LogCS);
    
  UnsetROAttrib(sIniFilePath);
  SavefmMainSettings;

  slUsers.Free;
  slLog.Free;
end;

procedure TfmMain.LoadUserSettings;
var
  ifMain: TIniFile;
  Cnt: Integer;
  CurrentDir, ts: String;
begin
  CurrentDir:=GetCurrentDir;
  try
    SetCurrentDir(sExeDirPath);

    if FileExists(sIniFilePath) then
      begin
        ifMain:=TIniFile.Create(sIniFilePath);
        ifMain.UpdateFile;

        bAllowAnonymousLogin:=ifMain.ReadBool(sApplicationIniSection, 'AllowAnonymousLogin', true);
        sAnonymousDir:=ExpandFileName(ifMain.ReadString(sApplicationIniSection, 'AnonymousDir', ''));
        if sAnonymousDir = '' then
          sAnonymousDir:=IncludeTrailingPathDelimiter(sExeDirPath)+'public';
        bCheckPassForValidMail:=ifMain.ReadBool(sApplicationIniSection, 'CheckPassForValidMail', true);

        ifMain.Free;
      end
    else
      begin
        bAllowAnonymousLogin:=true;
        sAnonymousDir:=IncludeTrailingPathDelimiter(sExeDirPath)+'public';
        bCheckPassForValidMail:=true;
      end;

    try
      slUsers.Clear;
      FileMode:=fmOpenRead;
      Cnt:=0;
      AssignFile(Input, IncludeTrailingPathDelimiter(sExeDirPath)+'users.txt');
      try
        while not EOF(Input) do
          begin
            ReadLn(ts);
            if Cnt mod 3 = 1 then
              slUsers.Add(ExpandFileName(ts))
            else
              slUsers.Add(ts);
            Inc(Cnt);
          end;
      finally
        CloseFile(Input);
      end;
    except
    end;
  finally
    SetCurrentDir(CurrentDir);
  end;
end;

procedure TfmMain.SaveUserSettings;
var
  ifMain: TIniFile;
  i: Integer;
  ts: String;
begin
  try
    ifMain:=TIniFile.Create(sIniFilePath);

    ifMain.WriteBool(sApplicationIniSection, 'AllowAnonymousLogin', fmUserSettings.cbAllowAnonymousLogin.Checked);
    ts:=ExtractRelativePath(Application.ExeName, IncludeTrailingPathDelimiter(fmUserSettings.edAnonymousDir.Text));
    if ts = '' then
      ts:='.\';
    ifMain.WriteString(sApplicationIniSection, 'AnonymousDir', ts);
    ifMain.WriteBool(sApplicationIniSection, 'CheckPassForValidMail', fmUserSettings.cbCheckPassForValidMail.Checked);

    ifMain.Free;
  except
  end;

  bAllowAnonymousLogin:=fmUserSettings.cbAllowAnonymousLogin.Checked;
  sAnonymousDir:=IncludeTrailingPathDelimiter(fmUserSettings.edAnonymousDir.Text);
  bCheckPassForValidMail:=fmUserSettings.cbCheckPassForValidMail.Checked;

  slUsers.Clear;
  for i:=0 to fmUserSettings.lvUsers.Items.Count-1 do
    begin
      slUsers.Add(fmUserSettings.lvUsers.Items[i].Caption);
      slUsers.Add(fmUserSettings.lvUsers.Items[i].SubItems[0]);
      slUsers.Add(fmUserSettings.lvUsers.Items[i].SubItems[1]);
    end;

  try
    AssignFile(Output, IncludeTrailingPathDelimiter(sExeDirPath)+'users.txt');
    try
      Rewrite(Output);
      for i:=0 to slUsers.Count-1 do
        if i mod 3 = 1 then
          WriteLn(ExtractRelativePath(Application.ExeName, slUsers[i]))
        else
          WriteLn(slUsers[i]);
    finally
      CloseFile(Output);
    end;
  except
  end;
end;

procedure TfmMain.ServerConnect(AThread: TIdPeerThread);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add(DateTimeToStr(Now) + ':   ' + AThread.Connection.Socket.Binding.PeerIP + '   ' + IntToStr(AThread.Connection.Socket.Binding.PeerPort));
  finally
    LeaveCriticalSection(LogCS);
  end;
  UserName:='';
  UserType:=utNone;
  Authenticated:=false;
  CurrentDir:='';
  PASV:=false;
  TransferType:=ttASCII;
  DataChannel:=nil;
end;

procedure TfmMain.EnableControls(bEnable: boolean);
begin
  laPort.Enabled:=bEnable;
  sedPort.Enabled:=bEnable;
  if bEnable then
    sedPort.Color:=clWindow
  else
    sedPort.Color:=clBtnFace;

  btUserSettings.Enabled:=bEnable;
end;

procedure TfmMain.sbtStartStopClick(Sender: TObject);
var
  Cursor: TCursor;
begin
  if sbtStartStop.Tag = 0 then
    begin
      Cursor:=Screen.Cursor;
      try
        Screen.Cursor:=crHourGlass;
        Server.DefaultPort:=sedPort.Value;
        Server.Active:=true;
        Timer.Enabled:=true;
        EnableControls(false);
        sbtStartStop.Glyph:=nil;
        ilButtons.GetBitmap(1, sbtStartStop.Glyph);
        sbtStartStop.Caption:='&';
        sbtStartStop.Tag:=1;
        EnterCriticalSection(LogCS);
        try
          slLog.Add(DateTimeToStr(Now) + ':     ' + IntToStr(sedPort.Value) + ' ');
        finally
          LeaveCriticalSection(LogCS);
        end;
      finally
        Screen.Cursor:=Cursor;
      end;
    end
  else
    begin
      Cursor:=Screen.Cursor;
      try
        Screen.Cursor:=crHourGlass;
        Server.Active:=false;
        Timer.Enabled:=false;
        EnableControls(true);
        sbtStartStop.Glyph:=nil;
        ilButtons.GetBitmap(0, sbtStartStop.Glyph);
        sbtStartStop.Caption:='&';
        sbtStartStop.Tag:=0;
        EnterCriticalSection(LogCS);
        try
          slLog.Add(DateTimeToStr(Now) + ':   ');
        finally
          LeaveCriticalSection(LogCS);
        end;
        Timer.OnTimer(Self); // carry out
      finally
        Screen.Cursor:=Cursor;
      end;
    end
end;

procedure TfmMain.TimerTimer(Sender: TObject);
begin
  EnterCriticalSection(LogCS);
  try
    memLog.Lines.AddStrings(slLog);
    slLog.Clear;
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServerDisconnect(AThread: TIdPeerThread);
begin
  if Assigned(DataChannel) then
    begin
      DataChannel.Free;
      DataChannel:=nil;
    end;

  EnterCriticalSection(LogCS);
  try
    slLog.Add(DateTimeToStr(Now) + ':  ' + AThread.Connection.Socket.Binding.PeerIP + '    ' + IntToStr(AThread.Connection.Socket.Binding.PeerPort));
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure LogOutcoming(s: String);
begin
  if (Length(s) > 1) and (s[Length(s)-1] = #13) then
    Delete(s, Length(s)-1, 2);
  slLog.Add('> '+s);
end;

procedure TfmMain.ServeridcUSERCommand(ASender: TIdCommand);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if (AnsiSameText(ASender.UnparsedParams, 'anonymous') or AnsiSameText(ASender.UnparsedParams, 'ftp') or AnsiSameText(ASender.UnparsedParams, 'guest')) and bAllowAnonymousLogin then
    begin
      UserName:=ASender.UnparsedParams;
      UserType:=utAnonymous;
      ASender.Reply.SetReply(331, 'Anonymous login OK, send your complete email address as a password');

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
    end
  else
    if Length(ASender.UnparsedParams) > 0 then
      begin
        UserName:=ASender.UnparsedParams;
        UserType:=utRegisteredUser;
        ASender.Reply.SetReply(331, 'User name okay, need password');

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
       end
    else
      begin
        UserName:='';
        UserType:=utNone;
        ASender.Reply.NumericCode:=332;
        Server.ReplyTexts.UpdateText(ASender.Reply);

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
      end;
end;

function CheckMailAddress(const sAddr: String): boolean;
var
  nAtPos, nDotPos: Integer;
begin
  nAtPos:=Pos('@', sAddr);
  nDotPos:=Pos('.', sAddr);
  Result:=(nAtPos <> 0) and (nDotPos <> 0) and (nDotPos - nAtPos > 1) and (nAtPos < Length(sAddr));
end;

procedure TfmMain.ServeridcPASSCommand(ASender: TIdCommand);
var
  bOK: boolean;
  i, Count: Integer;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  case UserType of
    utAnonymous:
      begin
        bOK:=Length(ASender.UnparsedParams) > 0;
        if bOK and bCheckPassForValidMail then
          bOK:=CheckMailAddress(ASender.UnparsedParams);
        if bOK then
          begin
            Authenticated:=true;
            Password:=ASender.UnparsedParams;
            ASender.Reply.SetReply(230, 'Anonymous user logged in, proceed');
          end
        else
          begin
            Authenticated:=false;
            Password:='';
            UserName:='';
            UserType:=utNone;
            ASender.Reply.NumericCode:=530;
            Server.ReplyTexts.UpdateText(ASender.Reply);
          end;

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
      end;
    utRegisteredUser:
      begin
        bOK:=false;
        Count:=slUsers.Count div 3;
        for i:=0 to Count - 1 do
          if slUsers[i*3] = UserName then
            if slUsers[i*3+2] = MD5StrToString(ASender.UnparsedParams) then
              begin
                bOK:=true;
                Break;
              end
            else
              Break;
        if bOK then
          begin
            Authenticated:=true;
            Password:=ASender.UnparsedParams;
            ASender.Reply.SetReply(230, 'User logged in, proceed');
          end
        else
          begin
            Authenticated:=false;
            Password:='';
            UserName:='';
            UserType:=utNone;
            ASender.Reply.NumericCode:=530;
            Server.ReplyTexts.UpdateText(ASender.Reply);
          end;

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
      end;
  else
    ASender.Reply.NumericCode:=503;
    Server.ReplyTexts.UpdateText(ASender.Reply);

    EnterCriticalSection(LogCS);
    try
      LogOutcoming(ASender.Reply.GenerateReply);
    finally
      LeaveCriticalSection(LogCS);
    end;
  end;
end;

function GetUserWorkingDir: String;
var
  i, Count: Integer;
begin
  if UserType = utAnonymous then
    Result:=sAnonymousDir
  else
    begin
      Count:=slUsers.Count div 3;
      for i:=0 to Count - 1 do
        if slUsers[i*3] = UserName then
          begin
            Result:=slUsers[i*3+1];
            Break;
          end;
    end;
end;

procedure TfmMain.ServeridcCWDCommand(ASender: TIdCommand);
var
  ts: String;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
      
      Exit;
    end;

  ts:=ASender.UnparsedParams;
  if (Length(ts) > 0) and (ts[Length(ts)] = '/') then
    Delete(ts, Length(ts), 1);

  if DirectoryExists(GetUserWorkingDir+StringReplace(ts, '/', '\', [rfReplaceAll])) then
    begin
      CurrentDir:=ts;
      ASender.Reply.NumericCode:=250;
      Server.ReplyTexts.UpdateText(ASender.Reply);
    end
  else
    ASender.Reply.SetReply(550, ASender.UnparsedParams+': no such file or directory');

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServerNoCommandHandler(ASender: TIdTCPServer;
  const AData: String; AThread: TIdPeerThread);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+AData);
  finally
    LeaveCriticalSection(LogCS);
  end;

  AThread.Connection.WriteLn('550 "'+AData+'" command not implemented');

  EnterCriticalSection(LogCS);
  try
    LogOutcoming('550 "'+AData+'" command not implemented');
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServeridcPWDCommand(ASender: TIdCommand);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
      
      Exit;
    end;

  if CurrentDir = '' then
    ASender.Reply.SetReply(257, '"/" is the current directory')
  else
    ASender.Reply.SetReply(257, '"'+CurrentDir+'" is the current directory');

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServeridcPASVCommand(ASender: TIdCommand);
var
  LParam: String;
  LBPort: Word;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
      
      Exit;
    end;

  DataChannel:=TIdSimpleServer.Create(nil);

  DataChannel.BoundIP:=TIdIOHandlerSocket(ASender.Thread.Connection.IOHandler).Binding.IP;
  DataChannel.BoundPort:=20;
  DataChannel.BeginListen;

  LBPort:=DataChannel.Binding.Port;
  LParam:=StringReplace(DataChannel.BoundIP, '.', ',', [rfReplaceAll]);
  LParam:=LParam + ',' + IntToStr(LBPort div 256) + ',' + IntToStr(LBPort mod 256);

  PASV:=true;
  ASender.Reply.SetReply(227, 'Entering Passive Mode (' + LParam + ')');

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure ListDirectory(const sPath: String; slList: TStringList);
var
  H: THandle;
  FindData: TWin32FindData;
  ts: String;
begin
  slList.Clear;

  H:=FindFirstFile(PChar(sPath+'*.*'), FindData);
  if H <> INVALID_HANDLE_VALUE then
    begin
      repeat
        if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0 then
          slList.Add('-rw-r--r--   1 kadyshev users    ' + Format('%:8d', [FindData.nFileSizeLow]) + ' Dec  1  1986 ' + StrPas(FindData.cFileName))
        else
          begin
            ts:=FindData.cFileName;
            if (ts <> '.') and (ts <> '..') then
              slList.Add('drw-r--r--   1 kadyshev users    ' + Format('%:8d', [512]) + ' Dec  1  1986 ' + StrPas(FindData.cFileName));
          end;
      until
        not FindNextFile(H, FindData);
      Windows.FindClose(H);
    end;
end;

procedure TfmMain.ServeridcLISTCommand(ASender: TIdCommand);
var
  slList: TStringList;
  DirPath: String;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
      
      Exit;
    end;

  DirPath:=IncludeTrailingPathDelimiter(GetUserWorkingDir+StringReplace(CurrentDir, '/', '\', [rfReplaceAll]));
    
  slList:=TStringList.Create;
  try
    ListDirectory(DirPath, slList);
    ASender.Reply.NumericCode:=125;
    ASender.SendReply;

    EnterCriticalSection(LogCS);
    try
      LogOutcoming(ASender.Reply.GenerateReply);
    finally
      LeaveCriticalSection(LogCS);
    end;
    
    try
      DataChannel.Listen;
      DataChannel.WriteStrings(slList);
      DataChannel.Disconnect;

      ASender.Reply.SetReply(226, 'Closing data connection');
      ASender.SendReply;

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
    except
      ASender.Reply.SetReply(426, 'Data connection closed abnormally');
      ASender.SendReply;      

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
    end;
  finally
    slList.Free;
    DataChannel.Free;
    DataChannel:=nil;
  end;
end;

procedure TfmMain.ServeridcSYSTCommand(ASender: TIdCommand);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;

      Exit;
    end;

  ASender.Reply.NumericCode:=215;
  Server.ReplyTexts.UpdateText(ASender.Reply);

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServeridcTYPECommand(ASender: TIdCommand);
var
  c: char;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;

      Exit;
    end;

  if Length(ASender.UnparsedParams)=1 then
    begin
      c:=UpCase(ASender.UnparsedParams[1]);

      case c of
        'A':
          begin
            TransferType:=ttASCII;
            ASender.Reply.SetReply(200, 'Type set to ' + c);
          end;
        'I':
          begin
            TransferType:=ttBinary;
            ASender.Reply.SetReply(200, 'Type set to ' + c);
          end;
      else
        ASender.Reply.SetReply(500, '"' + ASender.RawLine + '" not understood');
      end;
    end
  else
    ASender.Reply.SetReply(500, '"' + ASender.RawLine + '" not understood');

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;

procedure TfmMain.ServeridcRETRCommand(ASender: TIdCommand);
var
  Stream: TFileStream;
  DirPath: String;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;
      
      Exit;
    end;

  DirPath:=IncludeTrailingPathDelimiter(GetUserWorkingDir+StringReplace(CurrentDir, '/', '\', [rfReplaceAll]));

  try
    if FileExists(DirPath + ASender.UnparsedParams) then
      begin
        Stream:=TFileStream.Create(DirPath + ASender.UnparsedParams, fmOpenRead);
        try
          ASender.Reply.NumericCode:=150;
          ASender.SendReply;

          EnterCriticalSection(LogCS);
          try
            LogOutcoming(ASender.Reply.GenerateReply);
          finally
            LeaveCriticalSection(LogCS);
          end;

          try
            DataChannel.Listen;
            DataChannel.WriteStream(Stream);
            DataChannel.Disconnect;

            ASender.Reply.SetReply(226, 'Closing data connection');
            ASender.SendReply;
            
            EnterCriticalSection(LogCS);
            try
              LogOutcoming(ASender.Reply.GenerateReply);
            finally
              LeaveCriticalSection(LogCS);
            end;
          except
            ASender.Reply.SetReply(426, 'Data connection closed abnormally');
            ASender.SendReply;

            EnterCriticalSection(LogCS);
            try
              LogOutcoming(ASender.Reply.GenerateReply);
            finally
              LeaveCriticalSection(LogCS);
            end;
          end;
        finally
          Stream.Free;
        end;
      end
    else
      begin
        ASender.Reply.SetReply(550, ASender.UnparsedParams+': no such file or directory');

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
      end;
  finally
    DataChannel.Free;
    DataChannel:=nil;
  end;
end;

procedure TfmMain.ServeridcSTORCommand(ASender: TIdCommand);
var
  Stream: TMemoryStream;
  DirPath: String;
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  if not Authenticated then
    begin
      ASender.Reply.NumericCode:=530;
      Server.ReplyTexts.UpdateText(ASender.Reply);

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;

      Exit;
    end;

  DirPath:=IncludeTrailingPathDelimiter(GetUserWorkingDir+StringReplace(CurrentDir, '/', '\', [rfReplaceAll]));

  try
    Stream:=TMemoryStream.Create;
    try
      ASender.Reply.NumericCode:=150;
      ASender.SendReply;

      EnterCriticalSection(LogCS);
      try
        LogOutcoming(ASender.Reply.GenerateReply);
      finally
        LeaveCriticalSection(LogCS);
      end;

      try
        DataChannel.Listen;
        DataChannel.ReadStream(Stream, -1, true);
        DataChannel.Disconnect;

        try
          Stream.SaveToFile(DirPath + ASender.UnparsedParams);

          ASender.Reply.SetReply(226, 'Closing data connection');
          ASender.SendReply;

          EnterCriticalSection(LogCS);
          try
            LogOutcoming(ASender.Reply.GenerateReply);
          finally
            LeaveCriticalSection(LogCS);
          end;
        except
          ASender.Reply.SetReply(550, 'Requested action aborted: local error in processing');
          ASender.SendReply;

          EnterCriticalSection(LogCS);
          try
            LogOutcoming(ASender.Reply.GenerateReply);
          finally
            LeaveCriticalSection(LogCS);
          end;
        end;

      except
        ASender.Reply.SetReply(426, 'Data connection closed abnormally');
        ASender.SendReply;

        EnterCriticalSection(LogCS);
        try
          LogOutcoming(ASender.Reply.GenerateReply);
        finally
          LeaveCriticalSection(LogCS);
        end;
      end;
    finally
      Stream.Free;
    end;
  finally
    DataChannel.Free;
    DataChannel:=nil;
  end;
end;

procedure TfmMain.ServeridcQUITCommand(ASender: TIdCommand);
begin
  EnterCriticalSection(LogCS);
  try
    slLog.Add('< '+ASender.RawLine);
  finally
    LeaveCriticalSection(LogCS);
  end;

  EnterCriticalSection(LogCS);
  try
    LogOutcoming(ASender.Reply.GenerateReply);
  finally
    LeaveCriticalSection(LogCS);
  end;
end;
     
end.
